@@ -19,6 +19,7 @@
 
 #include "e-book-backend-file-index.h"
 #include "e-book-backend-file-log.h"
+#include "e-book-backend-file-utils.h"
 
 #include "libedataserver/e-sexp.h"
 #include "libebook/e-contact.h"
@@ -37,7 +38,7 @@ G_DEFINE_TYPE (EBookBackendFileIndex, e_book_backend_file_index, G_TYPE_OBJECT)
 
 /* structures used for the parsing side of things */
 
-typedef struct 
+typedef struct
 {
   gchar *name;            /* function name */
   ESExpFunc *test_function;   /* for just testing to see if we can use the index */
@@ -79,10 +80,10 @@ typedef enum
 
 typedef struct _EBookBackendFileIndexData EBookBackendFileIndexData;
 
-typedef void (*EBookBackendFileIndexAddFunc) (EBookBackendFileIndex *index, EContact *contact,
-    EBookBackendFileIndexData *data, DB *db, const gchar *id);
-typedef void (*EBookBackendFileIndexRemoveFunc) (EBookBackendFileIndex *index, EContact *contact,
-    EBookBackendFileIndexData *data, DB *db, const gchar *id);
+typedef int (*EBookBackendFileIndexAddFunc) (EBookBackendFileIndex *index, EContact *contact,
+    EBookBackendFileIndexData *data, DB *db, const gchar *id, GPtrArray *ops);
+typedef int (*EBookBackendFileIndexRemoveFunc) (EBookBackendFileIndex *index, EContact *contact,
+    EBookBackendFileIndexData *data, DB *db, const gchar *id, GPtrArray *ops);
 
 typedef int (*EBookBackendFileIndexOrderFunc) (DB *secondary, const DBT *akey, const DBT *bkey);
 
@@ -97,15 +98,15 @@ struct _EBookBackendFileIndexData
   EBookBackendFileIndexOrderFunc order_func; /* ordering func */
 };
 
-static void first_last_add_cb (EBookBackendFileIndex *index, EContact *contact, 
-    EBookBackendFileIndexData *data, DB *db, const gchar *uid);
-static void last_first_add_cb (EBookBackendFileIndex *index, EContact *contact, 
-    EBookBackendFileIndexData *data, DB *db, const gchar *uid);
+static int first_last_add_cb (EBookBackendFileIndex *index, EContact *contact,
+    EBookBackendFileIndexData *data, DB *db, const gchar *uid, GPtrArray *ops);
+static int last_first_add_cb (EBookBackendFileIndex *index, EContact *contact,
+    EBookBackendFileIndexData *data, DB *db, const gchar *uid, GPtrArray *ops);
 
-static void first_last_remove_cb (EBookBackendFileIndex *index, EContact *contact, 
-    EBookBackendFileIndexData *data, DB *db, const gchar *uid);
-static void last_first_remove_cb (EBookBackendFileIndex *index, EContact *contact, 
-    EBookBackendFileIndexData *data, DB *db, const gchar *uid);
+static int first_last_remove_cb (EBookBackendFileIndex *index, EContact *contact,
+    EBookBackendFileIndexData *data, DB *db, const gchar *uid, GPtrArray *ops);
+static int last_first_remove_cb (EBookBackendFileIndex *index, EContact *contact,
+    EBookBackendFileIndexData *data, DB *db, const gchar *uid, GPtrArray *ops);
 
 static int lexical_ordering_cb (DB *secondary, const DBT *akey, const DBT *bkey);
 
@@ -126,16 +127,13 @@ struct _EBookBackendFileIndexPrivate {
   char *index_dirname; /* dirname where the index dbs are stored */
 };
 
-static gboolean index_populate (EBookBackendFileIndex *index, GList *fields);
-static void index_add_contact (EBookBackendFileIndex *index, EContact *contact, 
-    EBookBackendFileIndexData *data);
-static void index_remove_contact (EBookBackendFileIndex *index, EContact *contact, 
-    EBookBackendFileIndexData *data);
+static gboolean index_populate (EBookBackendFileIndex *index, GList *fields, GPtrArray *ops);
+static int index_add_contact (EBookBackendFileIndex *index, EContact *contact,
+    EBookBackendFileIndexData *data, GPtrArray *ops);
+static int index_remove_contact (EBookBackendFileIndex *index, EContact *contact,
+    EBookBackendFileIndexData *data, GPtrArray *ops);
 static void index_sync (EBookBackendFileIndex *index, const EBookBackendFileIndexData *data);
 static gboolean index_close_db_func (gpointer key, gpointer value, gpointer userdata);
-static void dbt_fill_with_string (DBT *dbt, gchar *str);
-
-#define E_BOOK_BACKEND_FILE_VERSION_NAME "PAS-DB-VERSION"
 
 static void
 e_book_backend_file_index_dispose (GObject *object)
@@ -146,7 +144,7 @@ e_book_backend_file_index_dispose (GObject *object)
   if (priv->sdbs)
   {
     /* close all the open index databases */
-    g_hash_table_foreach_remove (priv->sdbs, index_close_db_func, 
+    g_hash_table_foreach_remove (priv->sdbs, index_close_db_func,
         E_BOOK_BACKEND_FILE_INDEX (object));
 
     g_hash_table_unref (priv->sdbs);
@@ -220,7 +218,7 @@ e_book_backend_file_index_is_usable (EBookBackendFileIndex *index, const gchar *
    */
   for (i = 0; i < G_N_ELEMENTS (sexp_symbols); i++)
   {
-    e_sexp_add_function (sexp, 0, sexp_symbols[i].name, 
+    e_sexp_add_function (sexp, 0, sexp_symbols[i].name,
         sexp_symbols[i].test_function, &res);
   }
 
@@ -269,7 +267,7 @@ e_book_backend_file_index_query (EBookBackendFileIndex *index, const gchar *quer
    */
   for (i = 0; i < G_N_ELEMENTS (sexp_symbols); i++)
   {
-    e_sexp_add_function (sexp, 0, sexp_symbols[i].name, 
+    e_sexp_add_function (sexp, 0, sexp_symbols[i].name,
         sexp_symbols[i].result_function, index);
   }
 
@@ -399,7 +397,8 @@ e_book_backend_file_index_setup_indicies (EBookBackendFileIndex *index, DB *db,
 
     /* we cannot pass DB_CREATE in the first time, because if index dbs aren't
      * present we need to populate them */
-    db_error = sdb->open (sdb, NULL, index_db_path, NULL, DB_BTREE, DB_THREAD, 0666);
+    db_error = sdb->open (sdb, NULL, index_db_path, NULL, DB_BTREE,
+        DB_THREAD | DB_AUTO_COMMIT, 0666);
     if (db_error != 0)
     {
       DEBUG ("%s doesn't exist, try to create it", index_db_path);
@@ -413,7 +412,8 @@ e_book_backend_file_index_setup_indicies (EBookBackendFileIndex *index, DB *db,
         return FALSE;
       }
 
-      db_error = sdb->open (sdb, NULL, index_db_path, NULL, DB_BTREE, DB_CREATE | DB_THREAD, 0666);
+      db_error = sdb->open (sdb, NULL, index_db_path, NULL, DB_BTREE,
+          DB_CREATE | DB_THREAD | DB_AUTO_COMMIT, 0666);
       if (db_error != 0)
       {
         WARNING ("creating %s failed: %s", index_db_path, db_strerror (db_error));
@@ -435,14 +435,26 @@ e_book_backend_file_index_setup_indicies (EBookBackendFileIndex *index, DB *db,
   /* now go through and populate these */
   if (dbs_to_populate)
   {
-    if (!index_populate (index, dbs_to_populate))
+    GPtrArray *ops = g_ptr_array_new ();
+
+    if (!index_populate (index, dbs_to_populate, ops))
     {
       WARNING ("Error whilst trying to populate the index");
+      txn_ops_free (ops);
       g_list_free (dbs_to_populate);
       return FALSE;
     }
     g_list_free (dbs_to_populate);
 
+    /* create and commit the transaction */
+    db_error = txn_execute_ops (env, NULL, ops);
+    txn_ops_free (ops);
+
+    if (db_error != 0) {
+      WARNING ("cannot commit index populate transaction: %s", db_strerror (db_error));
+      return FALSE;
+    }
+
     /* sync the db file, since we created dbs here */
     e_book_backend_file_index_sync (index);
   }
@@ -450,51 +462,106 @@ e_book_backend_file_index_setup_indicies (EBookBackendFileIndex *index, DB *db,
   return TRUE;
 }
 
-void
-e_book_backend_file_index_add_contact (EBookBackendFileIndex *index, EContact *contact)
+int
+e_book_backend_file_index_add_contact (EBookBackendFileIndex *index, EContact *contact, GPtrArray *ops)
 {
   gint i = 0;
+  int retval = 0;
 
-  g_return_if_fail (E_IS_BOOK_BACKEND_FILE_INDEX (index));
-  g_return_if_fail (E_IS_CONTACT (contact));
+  g_return_val_if_fail (E_IS_BOOK_BACKEND_FILE_INDEX (index), EINVAL);
+  g_return_val_if_fail (E_IS_CONTACT (contact), EINVAL);
+  g_return_val_if_fail (ops, EINVAL);
 
   /* for each index database try add this contact */
   for (i = 0; i < G_N_ELEMENTS (indexes); i++)
   {
-    index_add_contact (index, contact, (EBookBackendFileIndexData *)&indexes[i]);
+    retval = index_add_contact (index, contact, (EBookBackendFileIndexData *)&indexes[i], ops);
+    if (retval != 0) {
+      break;
+    }
   }
+
+  return retval;
 }
 
-void
-e_book_backend_file_index_remove_contact (EBookBackendFileIndex *index, EContact *contact)
+int
+e_book_backend_file_index_remove_contact (EBookBackendFileIndex *index, EContact *contact, GPtrArray *ops)
 {
   gint i = 0;
+  int retval = 0;
 
-  g_return_if_fail (E_IS_BOOK_BACKEND_FILE_INDEX (index));
-  g_return_if_fail (E_IS_CONTACT (contact));
+  g_return_val_if_fail (E_IS_BOOK_BACKEND_FILE_INDEX (index), EINVAL);
+  g_return_val_if_fail (E_IS_CONTACT (contact), EINVAL);
+  g_return_val_if_fail (ops, EINVAL);
 
   /* for each index database try add this contact */
   for (i = 0; i < G_N_ELEMENTS (indexes); i++)
   {
-    index_remove_contact (index, contact, (EBookBackendFileIndexData *)&indexes[i]);
+    retval = index_remove_contact (index, contact, (EBookBackendFileIndexData *)&indexes[i], ops);
+    if (retval != 0) {
+      break;
+    }
   }
+
+  return retval;
 }
 
-void
-e_book_backend_file_index_modify_contact (EBookBackendFileIndex *index, EContact *old_contact, EContact *new_contact)
+int
+e_book_backend_file_index_modify_contact (EBookBackendFileIndex *index, EContact *old_contact, EContact *new_contact, GPtrArray *ops)
 {
   gint i = 0;
+  int retval = 0;
 
-  g_return_if_fail (E_IS_BOOK_BACKEND_FILE_INDEX (index));
-  g_return_if_fail (E_IS_CONTACT (old_contact));
-  g_return_if_fail (E_IS_CONTACT (new_contact));
+  g_return_val_if_fail (E_IS_BOOK_BACKEND_FILE_INDEX (index), EINVAL);
+  g_return_val_if_fail (E_IS_CONTACT (old_contact), EINVAL);
+  g_return_val_if_fail (E_IS_CONTACT (new_contact), EINVAL);
+  g_return_val_if_fail (ops, EINVAL);
 
   /* for each index database try add this contact */
   for (i = 0; i < G_N_ELEMENTS (indexes); i++)
   {
-    index_remove_contact (index, old_contact, (EBookBackendFileIndexData *)&indexes[i]);
-    index_add_contact (index, new_contact, (EBookBackendFileIndexData *)&indexes[i]);
+    retval = index_remove_contact (index, old_contact,
+                    (EBookBackendFileIndexData *)&indexes[i], ops);
+    if (retval != 0) {
+      break;
+    }
+
+    retval = index_add_contact (index, new_contact,
+                    (EBookBackendFileIndexData *)&indexes[i], ops);
+    if (retval != 0) {
+      break;
+    }
   }
+
+  return retval;
+}
+
+int
+e_book_backend_file_index_truncate (EBookBackendFileIndex *index)
+{
+  int db_error = 0;
+  EBookBackendFileIndexPrivate *priv;
+  GHashTableIter iter;
+  gpointer key, value;
+
+  g_return_val_if_fail (E_IS_BOOK_BACKEND_FILE_INDEX (index), EINVAL);
+
+  priv = GET_PRIVATE (index);
+
+  g_hash_table_iter_init (&iter, priv->sdbs);
+  while (g_hash_table_iter_next (&iter, &key, &value)) {
+    /* truncate db, escape on failure */
+    u_int32_t count;
+    DB *db = value;
+
+    db_error = db->truncate (db, NULL, &count, DB_AUTO_COMMIT);
+    if (db_error != 0) {
+      WARNING ("db->truncate failed: %s", db_strerror (db_error));
+      return db_error;
+    }
+  }
+
+  return db_error;
 }
 
 void
@@ -557,7 +624,7 @@ e_book_backend_file_index_get_ordered_ids (EBookBackendFileIndex *index, const g
 
   if (db_error != 0)
   {
-    WARNING ("db->cursor failed: %s", db_strerror (db_error)); 
+    WARNING ("db->cursor failed: %s", db_strerror (db_error));
     return NULL;
   }
 
@@ -576,14 +643,14 @@ e_book_backend_file_index_get_ordered_ids (EBookBackendFileIndex *index, const g
 
   if (db_error != DB_NOTFOUND)
   {
-    g_warning (G_STRLOC ":dbc->c_get failed: %s", db_strerror (db_error));
+    WARNING ("dbc->c_get failed: %s", db_strerror (db_error));
   }
 
   db_error = dbc->c_close (dbc);
 
   if (db_error != 0)
   {
-    g_warning (G_STRLOC ":dbc->c_close failed: %s", db_strerror (db_error));
+    WARNING ("dbc->c_close failed: %s", db_strerror (db_error));
   }
 
   return ids;
@@ -632,7 +699,7 @@ test_always_fails (ESExp *sexp, gint argc, ESExpResult **argv, gpointer userdata
   return result;
 }
 
-/* 
+/*
  * return true or false based on whether we recognise that query field as
  * something we have an index for or not
  */
@@ -713,22 +780,12 @@ test_generic_field_is_suffix_indexed_vcard (ESExp *sexp, gint argc, ESExpResult
 
 /* index operations themselves */
 
-/* boring helper function */
-static void
-dbt_fill_with_string (DBT *dbt, gchar *str)
-{
-  memset (dbt, 0, sizeof (DBT));
-  dbt->data = (void *)str;
-  dbt->size = strlen (str) + 1;
-  dbt->flags = DB_DBT_USERMEM;
-}
-
 /* when given a list of fields, walk through the primary database, walk
  * through all contacts, parse them, then call the function to add to each
  * index that needs updating
  */
-static gboolean 
-index_populate (EBookBackendFileIndex *index, GList *fields)
+static gboolean
+index_populate (EBookBackendFileIndex *index, GList *fields, GPtrArray *ops)
 {
   EBookBackendFileIndexPrivate *priv = GET_PRIVATE (index);
 
@@ -745,7 +802,7 @@ index_populate (EBookBackendFileIndex *index, GList *fields)
 
   if (db_error != 0)
   {
-    WARNING ("db->cursor failed: %s", db_strerror (db_error)); 
+    WARNING ("db->cursor failed: %s", db_strerror (db_error));
     return FALSE;
   }
 
@@ -756,19 +813,21 @@ index_populate (EBookBackendFileIndex *index, GList *fields)
 
   while (db_error == 0)
   {
-    if (!g_str_equal (id_dbt.data, E_BOOK_BACKEND_FILE_VERSION_NAME))
-    {
-      /* parse the vcard */
-      contact = e_contact_new_from_vcard (vcard_dbt.data);
+    /* parse the vcard */
+    contact = e_contact_new_from_vcard (vcard_dbt.data);
 
-      if (contact)
+    if (contact)
+    {
+      /* now we interate through all the indexes that need doing */
+      for (l = fields; l != NULL; l = l->next)
       {
-        /* now we interate through all the indexes that need doing */
-        for (l = fields; l != NULL; l = l->next)
-        {
-          index_add_contact (index, contact, (EBookBackendFileIndexData *)l->data);
+        if (0 != index_add_contact (index, contact, (EBookBackendFileIndexData *)l->data, ops)) {
+          WARNING ("cannot populate index");
+          g_object_unref (contact);
+          return FALSE;
         }
       }
+      g_object_unref (contact);
     }
 
     db_error = dbc->c_get (dbc, &id_dbt, &vcard_dbt, DB_NEXT);
@@ -776,7 +835,7 @@ index_populate (EBookBackendFileIndex *index, GList *fields)
 
   if (db_error != DB_NOTFOUND)
   {
-    g_warning (G_STRLOC ":dbc->c_get failed: %s", db_strerror (db_error));
+    WARNING ("dbc->c_get failed: %s", db_strerror (db_error));
     return FALSE;
   }
 
@@ -784,7 +843,7 @@ index_populate (EBookBackendFileIndex *index, GList *fields)
 
   if (db_error != 0)
   {
-    g_warning (G_STRLOC ":dbc->c_close failed: %s", db_strerror (db_error));
+    WARNING ("dbc->c_close failed: %s", db_strerror (db_error));
     return FALSE;
   }
 
@@ -826,7 +885,7 @@ get_key_attribute (EContact *contact)
     if (g_str_equal (attr_name, key_priority[0])) {
       if (!e_vcard_attribute_get_values (attr))
         continue;
-      
+
       g_hash_table_destroy (ht);
       return attr;
     }
@@ -860,7 +919,7 @@ get_key_attribute (EContact *contact)
   return key_attr;
 }
 
-enum 
+enum
 {
   ORDER_FIRSTLAST = 0,
   ORDER_LASTFIRST
@@ -904,127 +963,100 @@ get_key (EContact *contact, gint ordering)
   return key;
 }
 
-static void
+static int
 generic_name_add_cb (EBookBackendFileIndex *index, EContact *contact,
-    EBookBackendFileIndexData *data, DB *db, const gchar *uid, gint ordering)
+    EBookBackendFileIndexData *data, DB *db, const gchar *uid, gint ordering, GPtrArray *ops)
 {
-  DBT index_dbt, id_dbt;
   int db_error;
   char *key;
 
   key = get_key (contact, ordering);
   if (!key) {
     WARNING ("cannot create a sufficient key value");
-    return;
+    return EINVAL;
   }
-  dbt_fill_with_string (&index_dbt, key);
 
-  dbt_fill_with_string (&id_dbt, (gchar *)uid);
-
-  DEBUG ("adding to index with key %s and data %s",
-           (gchar *)index_dbt.data, (gchar *)id_dbt.data);
-
-  db_error = db->put (db, NULL, &index_dbt, &id_dbt, 0);
-  g_free (key);
+  DEBUG ("adding to index with key %s and data %s", key, uid);
 
+  db_error = txn_ops_add_new (ops, TXN_PUT, db, key, g_strdup (uid));
   if (db_error != 0) {
-    WARNING ("db->put failed: %s", db_strerror (db_error));
+    WARNING ("cannot create new txn item: %s", db_strerror (db_error));
   }
+
+  return db_error;
 }
 
-static void
+static int
 generic_name_remove_cb (EBookBackendFileIndex *index, EContact *contact,
-    EBookBackendFileIndexData *data, DB *db, const gchar *uid, gint ordering)
+    EBookBackendFileIndexData *data, DB *db, const gchar *uid, gint ordering, GPtrArray *ops)
 {
-  DBC *dbc = NULL; /* cursor */
-  DBT index_dbt, id_dbt;
   int db_error;
   char *key;
 
-  db_error = db->cursor (db, NULL, &dbc, 0);
-  if (db_error) {
-    WARNING ("db->cursor failed: %s", db_strerror (db_error));
-    return;
-  }
-
   key = get_key (contact, ordering);
   if (!key) {
     /* TODO: remove index only by uid */
-    g_warning ("cannot determine the key value");
-    return;
+    WARNING ("cannot determine the key value");
+    return EINVAL;
   }
-  dbt_fill_with_string (&index_dbt, key);
 
-  dbt_fill_with_string (&id_dbt, (gchar *)uid);
+  DEBUG ("removing from index with key %s and data %s", key, uid);
 
-  DEBUG ("removing from index with key %s and data %s",
-           (gchar *)index_dbt.data, (gchar *)id_dbt.data);
-
-  id_dbt.flags = 0;
-  index_dbt.flags = 0;
-  db_error = dbc->c_get (dbc, &index_dbt, &id_dbt, DB_GET_BOTH);
-  if (db_error) {
-    WARNING ("dbc->c_get failed: %s", db_strerror (db_error));
-  } else {
-    db_error = dbc->c_del (dbc, 0);
-    if (db_error != 0) {
-      WARNING ("dbc->c_del failed: %s", db_strerror (db_error));
-    }
+  db_error = txn_ops_add_new (ops, TXN_CDEL, db, key, g_strdup (uid));
+  if (db_error != 0) {
+    WARNING ("cannot create new txn item: %s", db_strerror (db_error));
   }
-  g_free (key);
 
-  db_error = dbc->c_close (dbc);
-  if (db_error) {
-    WARNING ("db->c_close failed: %s", db_strerror (db_error));
-  }
+  return db_error;
 }
 
-static void 
-first_last_add_cb (EBookBackendFileIndex *index, EContact *contact, 
-    EBookBackendFileIndexData *data, DB *db, const gchar *uid)
+static int
+first_last_add_cb (EBookBackendFileIndex *index, EContact *contact,
+    EBookBackendFileIndexData *data, DB *db, const gchar *uid, GPtrArray *ops)
 {
-  generic_name_add_cb (index, contact, data, db, uid, ORDER_FIRSTLAST);
+  return generic_name_add_cb (index, contact, data, db, uid, ORDER_FIRSTLAST, ops);
 }
 
-static void 
-first_last_remove_cb (EBookBackendFileIndex *index, EContact *contact, 
-    EBookBackendFileIndexData *data, DB *db, const gchar *uid)
+static int
+first_last_remove_cb (EBookBackendFileIndex *index, EContact *contact,
+    EBookBackendFileIndexData *data, DB *db, const gchar *uid, GPtrArray *ops)
 {
-  generic_name_remove_cb (index, contact, data, db, uid, ORDER_FIRSTLAST);
+  return generic_name_remove_cb (index, contact, data, db, uid, ORDER_FIRSTLAST, ops);
 }
 
-static void
-last_first_add_cb (EBookBackendFileIndex *index, EContact *contact, 
-    EBookBackendFileIndexData *data, DB *db, const gchar *uid)
+static int
+last_first_add_cb (EBookBackendFileIndex *index, EContact *contact,
+    EBookBackendFileIndexData *data, DB *db, const gchar *uid, GPtrArray *ops)
 {
-  generic_name_add_cb (index, contact, data, db, uid, ORDER_LASTFIRST);
+  return generic_name_add_cb (index, contact, data, db, uid, ORDER_LASTFIRST, ops);
 }
 
-static void
-last_first_remove_cb (EBookBackendFileIndex *index, EContact *contact, 
-    EBookBackendFileIndexData *data, DB *db, const gchar *uid)
+static int
+last_first_remove_cb (EBookBackendFileIndex *index, EContact *contact,
+    EBookBackendFileIndexData *data, DB *db, const gchar *uid, GPtrArray *ops)
 {
-  generic_name_remove_cb (index, contact, data, db, uid, ORDER_LASTFIRST);
+  return generic_name_remove_cb (index, contact, data, db, uid, ORDER_LASTFIRST, ops);
 }
 
-/* 
+/*
  * generic version of function to get the index key from a contact based on
  * the VCard field
  */
-static void
+static int
 generic_field_add (EBookBackendFileIndex *index, EContact *contact,
-    EBookBackendFileIndexData *data, DB *db, const gchar *uid)
+    EBookBackendFileIndexData *data, DB *db, const gchar *uid, GPtrArray *ops)
 {
   gchar *tmp = NULL;
-  DBT index_dbt, id_dbt;
   int db_error = 0;
   GList *attrs = NULL;
   GList *values = NULL;
+  GHashTable *ht;
+  GHashTableIter iter;
+  gpointer key, value;
 
-  dbt_fill_with_string (&id_dbt, (gchar *)uid);
-  
-  for (attrs = e_vcard_get_attributes (E_VCARD (contact)); 
-      attrs != NULL; 
+  ht = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+  for (attrs = e_vcard_get_attributes (E_VCARD (contact));
+      attrs != NULL;
       attrs = attrs->next)
   {
     EVCardAttribute *attr = (EVCardAttribute *)attrs->data;
@@ -1032,7 +1064,7 @@ generic_field_add (EBookBackendFileIndex *index, EContact *contact,
     if (g_str_equal (e_vcard_attribute_get_name (attr), data->vfield))
     {
       for (values = e_vcard_attribute_get_values (attr);
-          values != NULL; 
+          values != NULL;
           values = values->next)
       {
         /* normalize the phone number before we add it to db */
@@ -1053,46 +1085,46 @@ generic_field_add (EBookBackendFileIndex *index, EContact *contact,
           g_free (tmp);
           tmp = reversed;
         }
-        dbt_fill_with_string (&index_dbt, g_strstrip (tmp));
 
-        DEBUG ("adding to index '%s' with key %s and data %s", data->index_name,
-            (gchar *)index_dbt.data, (gchar *)id_dbt.data);
-
-        db_error = db->put (db, NULL, &index_dbt, &id_dbt, 0);
-        g_free (tmp);
-
-        if (db_error != 0)
-        {
-          WARNING ("db->put failed: %s", db_strerror (db_error));
-        }
+        g_strstrip (tmp);
+        g_hash_table_insert (ht, tmp, NULL);
       }
     }
   }
+
+  g_hash_table_iter_init (&iter, ht);
+  while (g_hash_table_iter_next (&iter, &key, &value))
+  {
+    DEBUG ("adding to index '%s' with key %s and data %s",
+           data->index_name, key, uid);
+
+    db_error = txn_ops_add_new (ops, TXN_PUT, db, g_strdup (key), g_strdup (uid));
+    if (db_error != 0)
+    {
+      WARNING ("cannot create new txn item: %s", db_strerror (db_error));
+      return db_error;
+    }
+  }
+  g_hash_table_destroy (ht);
+
+  return 0;
 }
 
-static void
+static int
 generic_field_remove (EBookBackendFileIndex *index, EContact *contact,
-    EBookBackendFileIndexData *data, DB *db, const gchar *uid)
+    EBookBackendFileIndexData *data, DB *db, const gchar *uid, GPtrArray *ops)
 {
-  DBT index_dbt, id_dbt;
-  GList *attrs; 
+  GList *attrs;
   GList *values;
   gint db_error = 0;
   gchar *tmp = NULL;
-  DBC *dbc = NULL; /* cursor */
+  GHashTable *ht;
+  GHashTableIter iter;
+  gpointer key, value;
 
-  /* create the cursor for the interation */
-  db_error = db->cursor (db, NULL, &dbc, 0);
-
-  if (db_error != 0)
-  {
-    WARNING ("db->cursor failed: %s", db_strerror (db_error)); 
-    return;
-  }
-
-  dbt_fill_with_string (&id_dbt, (gchar*)uid);
-  for (attrs = e_vcard_get_attributes (E_VCARD (contact)); 
-      attrs != NULL; 
+  ht = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+  for (attrs = e_vcard_get_attributes (E_VCARD (contact));
+      attrs != NULL;
       attrs = attrs->next)
   {
     EVCardAttribute *attr = (EVCardAttribute *)attrs->data;
@@ -1100,7 +1132,7 @@ generic_field_remove (EBookBackendFileIndex *index, EContact *contact,
     if (g_str_equal (e_vcard_attribute_get_name (attr), data->vfield))
     {
       for (values = e_vcard_attribute_get_values (attr);
-          values != NULL; 
+          values != NULL;
           values = values->next)
       {
         /* normalize phone number before we remove it, otherwise we won't find it */
@@ -1121,46 +1153,41 @@ generic_field_remove (EBookBackendFileIndex *index, EContact *contact,
           g_free (tmp);
           tmp = reversed;
         }
-        dbt_fill_with_string (&index_dbt, g_strstrip (tmp));
 
-        DEBUG ("removing from index '%s' with key %s and data %s", data->index_name,
-            (gchar *)index_dbt.data, (gchar *)id_dbt.data);
-
-        id_dbt.flags = 0;
-        index_dbt.flags = 0;
-        db_error = dbc->c_get (dbc, &index_dbt, &id_dbt, DB_GET_BOTH);
-        if (db_error != 0)
-        {
-          WARNING ("dbc->c_get failed: %s", db_strerror (db_error));
-        } else {
-          db_error = dbc->c_del (dbc, 0);
-          if (db_error != 0)
-          {
-            WARNING ("dbc->c_del failed: %s", db_strerror (db_error));
-          }
-        }
-        g_free (tmp);
+        g_strstrip (tmp);
+        g_hash_table_insert (ht, tmp, NULL);
       }
     }
   }
-  db_error = dbc->c_close (dbc);
 
-  if (db_error != 0)
+  g_hash_table_iter_init (&iter, ht);
+  while (g_hash_table_iter_next (&iter, &key, &value))
   {
-    WARNING ("db->c_closed failed: %s", db_strerror (db_error));
+    DEBUG ("removing from index '%s' with key %s and data %s",
+           data->index_name, key, uid);
+
+    db_error = txn_ops_add_new (ops, TXN_CDEL, db, g_strdup (key), g_strdup (uid));
+    if (db_error != 0) {
+      WARNING ("cannot create new txn item: %s", db_strerror (db_error));
+      return db_error;
+    }
   }
+  g_hash_table_destroy (ht);
+
+  return 0;
 }
 
-/* 
+/*
  * for a given index remove this contact
  */
-static void
-index_remove_contact (EBookBackendFileIndex *index, EContact *contact, 
-    EBookBackendFileIndexData *data)
+static int
+index_remove_contact (EBookBackendFileIndex *index, EContact *contact,
+    EBookBackendFileIndexData *data, GPtrArray *ops)
 {
   EBookBackendFileIndexPrivate *priv = GET_PRIVATE (index);
   gchar *uid = NULL;
   DB *db = NULL;
+  int retval;
 
   /* we need the uid, so that we can match both key and value of index-entries */
   uid = (gchar *)e_contact_get (contact, E_CONTACT_UID);
@@ -1170,12 +1197,14 @@ index_remove_contact (EBookBackendFileIndex *index, EContact *contact,
 
   if (data->contact_remove_func)
   {
-    data->contact_remove_func (index, contact, data, db, uid);
+    retval = data->contact_remove_func (index, contact, data, db, uid, ops);
   } else {
-    generic_field_remove (index, contact, data, db, uid);
+    retval = generic_field_remove (index, contact, data, db, uid, ops);
   }
 
   g_free (uid);
+
+  return retval;
 }
 
 /*
@@ -1183,13 +1212,14 @@ index_remove_contact (EBookBackendFileIndex *index, EContact *contact,
  * just add a single entry to the index database. for the multiple case we add
  * all the values for the field
  */
-static void
-index_add_contact (EBookBackendFileIndex *index, EContact *contact, 
-    EBookBackendFileIndexData *data)
+static int
+index_add_contact (EBookBackendFileIndex *index, EContact *contact,
+    EBookBackendFileIndexData *data, GPtrArray *ops)
 {
   EBookBackendFileIndexPrivate *priv = GET_PRIVATE (index);
   gchar *uid = NULL;
   DB *db = NULL;
+  int retval = 0;
 
   /* we need the uid, this becomes the data for our index databases */
   uid = (gchar *)e_contact_get (contact, E_CONTACT_UID);
@@ -1199,11 +1229,13 @@ index_add_contact (EBookBackendFileIndex *index, EContact *contact,
 
   if (data->contact_add_func)
   {
-    data->contact_add_func (index, contact, data, db, uid);
+    retval = data->contact_add_func (index, contact, data, db, uid, ops);
   } else {
-    generic_field_add (index, contact, data, db, uid);
+    retval = generic_field_add (index, contact, data, db, uid, ops);
   }
   g_free (uid);
+
+  return retval;
 }
 
 static void
@@ -1234,7 +1266,7 @@ find_query_term_for_vfield (const char *vfield)
       return indexes[i].query_term;
     }
   }
-  g_critical ("vcard-attribute %s is not indexed", vfield);
+  CRITICAL ("vcard-attribute %s is not indexed", vfield);
 
   return NULL;
 }
@@ -1244,7 +1276,7 @@ real_is_query (gboolean vcard_query, ESExp *sexp, gint argc, ESExpResult **argv,
 {
   EBookBackendFileIndex *index = (EBookBackendFileIndex *)userdata;
   EBookBackendFileIndexPrivate *priv = GET_PRIVATE (index);
-  
+
   ESExpResult *result = NULL;
   DBT index_dbt, id_dbt;
   DB *db = NULL;
@@ -1282,7 +1314,7 @@ real_is_query (gboolean vcard_query, ESExp *sexp, gint argc, ESExpResult **argv,
 
     if (db_error != 0)
     {
-      WARNING ("db->cursor failed: %s", db_strerror (db_error)); 
+      WARNING ("db->cursor failed: %s", db_strerror (db_error));
       goto out;
     } else {
       db_error = dbc->c_get (dbc, &index_dbt, &id_dbt, DB_SET);
@@ -1342,7 +1374,7 @@ real_query (gboolean vcard_query, gboolean endswith_query, ESExp *sexp, gint arg
 {
   EBookBackendFileIndex *index = (EBookBackendFileIndex *)userdata;
   EBookBackendFileIndexPrivate *priv = GET_PRIVATE (index);
-  
+
   ESExpResult *result = NULL;
   DBT index_dbt, id_dbt;
   DB *db = NULL;
@@ -1402,7 +1434,7 @@ real_query (gboolean vcard_query, gboolean endswith_query, ESExp *sexp, gint arg
 
     if (db_error != 0)
     {
-      WARNING ("db->cursor failed: %s", db_strerror (db_error)); 
+      WARNING ("db->cursor failed: %s", db_strerror (db_error));
       goto out;
     } else {
       db_error = dbc->c_get (dbc, &index_dbt, &id_dbt, DB_SET_RANGE);
@@ -1411,7 +1443,7 @@ real_query (gboolean vcard_query, gboolean endswith_query, ESExp *sexp, gint arg
       {
         if (g_str_has_prefix (index_dbt.data, query_key))
         {
-          DEBUG ("index query found: %s for %s for %s", 
+          DEBUG ("index query found: %s for %s for %s",
               (gchar *)id_dbt.data, query_key, (gchar *)index_dbt.data);
           g_ptr_array_add (ids, g_strdup (id_dbt.data));
           g_free (id_dbt.data);
@@ -1492,10 +1524,9 @@ index_close_db_func (gpointer key, gpointer value, gpointer userdata)
 }
 
 /* functions for the ordering */
-static int 
+static int
 lexical_ordering_cb (DB *secondary, const DBT *akey, const DBT *bkey)
 {
   return g_utf8_collate (akey->data, bkey->data);
 }
 
-
